#
# Implement an asyncio.Protocol for Electrum (clients)
#
#
import asyncio, json
import logging

logger = logging.getLogger(__name__)

class StratumProtocol(asyncio.Protocol):
    client = None
    closed = False
    transport = None
    buf = b""

    def connection_made(self, transport):
        self.transport = transport
        logger.debug("Transport connected ok")

    def connection_lost(self, exc):
        if not self.closed:
            self.closed = True
            self.close()
            self.client._connection_lost(self)

    def data_received(self, data):
        self.buf += data

        # Unframe the mesage. Expecting JSON.
        *lines, self.buf = self.buf.split(b'\n')

        for line in lines:
            if not line: continue

            try:
                msg = line.decode('utf-8', "error").strip()
            except UnicodeError as exc:
                logger.exception("Encoding issue on %r" % line)
                self.connection_lost(exc)
                return

            try:
                msg = json.loads(msg)
            except ValueError as exc:
                logger.exception("Bad JSON received from server: %r" % msg)
                self.connection_lost(exc)
                return

            #logger.debug("RX:\n%s", json.dumps(msg, indent=2))

            try:
                self.client._got_response(msg)
            except Exception as e:
                logger.exception("Trouble handling response! (%s)" % e)
                continue

    def send_data(self, message):
        '''
            Given an object, encode as JSON and transmit to the server.
        '''
        #logger.debug("TX:\n%s", json.dumps(message, indent=2))
        data = json.dumps(message).encode('utf-8') + b'\n'
        self.transport.write(data)

    def close(self):
        if not self.closed:
            try:
                self.transport.close()
            finally:
                self.closed = True

# EOF



#
# Client connect to an Electrum server.
#

# Runtime check for optional modules
from importlib import util as importutil
import json, warnings, asyncio, ssl
from .protocol import StratumProtocol

# Check if aiosocks is present, and load it if it is.
if importutil.find_spec("aiosocks") is not None:
    import aiosocks
    have_aiosocks = True
else:
    have_aiosocks = False

from collections import defaultdict
from .exc import ElectrumErrorResponse
import logging

logger = logging.getLogger(__name__)

class StratumClient:

    def __init__(self, loop=None):
        '''
            Setup state needed to handle req/resp from a single Stratum server.
            Requires a transport (TransportABC) object to do the communication.
        '''
        self.protocol = None

        self.next_id = 1
        self.inflight = {}
        self.subscriptions = defaultdict(list)

        # these are valid after connection
        self.server_version = None          # 'ElectrumX 1.13.0' or similar
        self.protocol_version = None        # float(1.4) or similar

        self.actual_connection = {}

        self.ka_task = None

        self.loop = loop or asyncio.get_event_loop()

        self.reconnect = None       # call connect() first

        # next step: call connect()

    def _connection_lost(self, protocol):
        # Ignore connection_lost for old connections
        if protocol is not self.protocol:
            return

        self.protocol = None
        logger.warn("Electrum server connection lost")

        # cleanup keep alive task
        if self.ka_task:
            self.ka_task.cancel()
            self.ka_task = None

    def close(self):
        if self.protocol:
            self.protocol.close()
            self.protocol = None
        if self.ka_task:
            self.ka_task.cancel()
            self.ka_task = None
            

    async def connect(self, server_info, proto_code=None, *,
                            use_tor=False, disable_cert_verify=False,
                            proxy=None, short_term=False):
        '''
            Start connection process.
            Destination must be specified in a ServerInfo() record (first arg).
        '''
        self.server_info = server_info
        if not proto_code:
             proto_code,*_ = server_info.protocols
        self.proto_code = proto_code

        logger.debug("Connecting to: %r" % server_info)

        if proto_code == 'g':       # websocket
            # to do this, we'll need a websockets implementation that
            # operates more like a asyncio.Transport
            # maybe: `asyncws` or `aiohttp` 
            raise NotImplementedError('sorry no WebSocket transport yet')

        hostname, port, use_ssl = server_info.get_port(proto_code)

        if use_tor:
            if have_aiosocks:
                # Connect via Tor proxy proxy, assumed to be on localhost:9050
                # unless a tuple is given with another host/port combo.
                try:
                    socks_host, socks_port = use_tor
                except TypeError:
                    socks_host, socks_port = 'localhost', 9050

                # basically no-one has .onion SSL certificates, and
                # pointless anyway.
                disable_cert_verify = True

                assert not proxy, "Sorry not yet supporting proxy->tor->dest"

                logger.debug(" .. using TOR")

                proxy = aiosocks.Socks5Addr(socks_host, int(socks_port))
            else:
                logger.debug("Error: want to use tor, but no aiosocks module.")

        if use_ssl == True and disable_cert_verify:
            # Create a more liberal SSL context that won't
            # object to self-signed certicates. This is 
            # very bad on public Internet, but probably ok
            # over Tor
            use_ssl = ssl.create_default_context()
            use_ssl.check_hostname = False
            use_ssl.verify_mode = ssl.CERT_NONE

            logger.debug(" .. SSL cert check disabled")

        async def _reconnect():
            if self.protocol: return        # race/duplicate work

            if proxy:
                if have_aiosocks:
                    transport, protocol = await aiosocks.create_connection(
                                            StratumProtocol, proxy=proxy,
                                            proxy_auth=None,
                                            remote_resolve=True, ssl=use_ssl,
                                            dst=(hostname, port))
                else:
                    logger.debug("Error: want to use proxy, but no aiosocks module.")
            else:
                transport, protocol = await self.loop.create_connection(
                                                        StratumProtocol, host=hostname,
                                                        port=port, ssl=use_ssl)

            self.protocol = protocol
            protocol.client = self

            # capture actual values used
            self.actual_connection = dict(hostname=hostname, port=int(port),
                                            ssl=bool(use_ssl), tor=bool(proxy))
            self.actual_connection['ip_addr'] = transport.get_extra_info('peername',
                                                        default=['unknown'])[0]

            if not short_term:
                self.ka_task = self.loop.create_task(self._keepalive())
            else:
                await self.get_server_version()
                logger.debug(f"Server version/protocol: {self.server_version} / {self.protocol_version}")

            logger.debug("Connected to: %r" % server_info)

        # close whatever we had
        if self.protocol:
            self.protocol.close()
            self.protocol = None

        self.reconnect = _reconnect
        await self.reconnect()

    async def get_server_version(self):
        # fetch version strings, save them
        self.server_version, pv = await self.RPC('server.version')
        self.protocol_version = float(pv)

    async def _keepalive(self):
        '''
            Keep our connect to server alive forever, with some 
            pointless traffic.
        '''
        while self.protocol:
            await self.get_server_version()
            logger.debug("Server version: " + repr(self.server_version))

            # Five minutes isn't really enough anymore; looks like
            # servers are killing 2-minute old idle connections now.
            # But decreasing interval this seems rude... actually   
            # much DDoS prevention in action now-a-days, so hard to workaround.
            await asyncio.sleep(120)


    def _send_request(self, method, params=[], is_subscribe = False):
        '''
            Send a new request to the server. Serialized the JSON and
            tracks id numbers and optional callbacks.
        '''
        # pick a new ID
        self.next_id += 1
        req_id = self.next_id

        # serialize as JSON
        msg = {'id': req_id, 'method': method, 'params': params}

        # subscriptions are a Q, normal requests are a future
        if is_subscribe:
            waitQ = asyncio.Queue()
            self.subscriptions[method].append(waitQ)

        fut = asyncio.Future(loop=self.loop)

        self.inflight[req_id] = (msg, fut)

        # send it via the transport, which serializes it
        if not self.protocol:
            logger.debug("Need to reconnect to server")

            async def connect_first():
                await self.reconnect()
                self.protocol.send_data(msg)

            self.loop.create_task(connect_first())
        else:
            # typical case, send request immediatedly, response is a future
            self.protocol.send_data(msg)

        return fut if not is_subscribe else (fut, waitQ)

    def _got_response(self, msg):
        '''
            Decode and dispatch responses from the server.

            Has already been unframed and deserialized into an object.
        '''

        #logger.debug("MSG: %r" % msg)

        resp_id = msg.get('id', None)

        if resp_id is None:
            # subscription traffic comes with method set, but no req id.
            method = msg.get('method', None)
            if not method:
                logger.error("Incoming server message had no ID nor method in it", msg)
                return

            # not obvious, but result is on params, not result, for subscriptions
            result = msg.get('params', None)

            logger.debug("Traffic on subscription: %s" % method)

            subs = self.subscriptions.get(method)
            for q in subs:
                self.loop.create_task(q.put(result))

            return

        assert 'method' not in msg
        result = msg.get('result')

        # fetch and forget about the request
        inf = self.inflight.pop(resp_id) 
        if not inf:
            logger.error("Incoming server message had unknown ID in it: %s" % resp_id)
            return

        # it's a future which is done now
        req, rv = inf

        if 'error' in msg:
            err = msg['error']
            
            logger.info("Error response: '%s'" % err)
            rv.set_exception(ElectrumErrorResponse(err, req))

        else:
            rv.set_result(result)

    def RPC(self, method, *params):
        '''
            Perform a remote command.

            Expects a method name, which look like:

                blockchain.address.get_balance

            .. and sometimes take arguments, all of which are positional.
    
            Returns a future which will you should await for
            the result from the server. Failures are returned as exceptions.
        '''
        assert '.' in method
        #assert not method.endswith('subscribe')

        if method.startswith('blockchain.address.') and self.protocol_version >= 1.3:
            # these methods have changed, but we can patch it.
            method, params = self.patch_addr_methods(method, params)

        return self._send_request(method, params)

    def patch_addr_methods(self, method, params):
        # blockchain.address.get_balance(addr) => blockchain.scripthash.get_balance(sh)
        from hashlib import sha256
        from binascii import b2a_hex
        try:
            from pycoin.symbols.btc import network as BTC       # bitcoin only!
        except ImportError:
            raise RuntimeError("we can patch obsolete protocol msgs, but need pycoin>=0.90")

        # convert from base58 into sha256(binary of script)?
        addr = BTC.parse(params[0])
        sh = sha256(addr.script()).digest()[::-1]

        return method.replace('.address.', '.scripthash.'), \
                    [str(b2a_hex(sh), 'ascii')]+list(params[1:])

    def subscribe(self, method, *params):
        '''
            Perform a remote command which will stream events/data to us.

            Expects a method name, which look like:
                server.peers.subscribe
            .. and sometimes take arguments, all of which are positional.
    
            Returns a tuple: (Future, asyncio.Queue).
            The future will have the result of the initial
            call, and the queue will receive additional
            responses as they happen.
        '''
        assert '.' in method
        assert method.endswith('subscribe')
        return self._send_request(method, params, is_subscribe=True)
        

if __name__ == '__main__':
    from transport import SocketTransport
    from svr_info import KnownServers, ServerInfo

    logging.basicConfig(format="%(asctime)-11s %(message)s", datefmt="[%d/%m/%Y-%H:%M:%S]")

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    proto_code = 's'

    if 0:
        ks = KnownServers()
        ks.from_json('servers.json')
        which = ks.select(proto_code, is_onion=True, min_prune=1000)[0]
    else:
        which = ServerInfo({
            "seen_at": 1465686119.022801,
            "ports": "t s",
            "nickname": "dunp",
            "pruning_limit": 10000,
            "version": "1.0",
            "hostname": "erbium1.sytes.net" })

    c = StratumClient(loop=loop)

    loop.run_until_complete(c.connect(which, proto_code, disable_cert_verify=True, use_tor=True))
    
    rv = loop.run_until_complete(c.RPC('server.peers.subscribe'))
    print("DONE!: this server has %d peers" % len(rv))
    loop.close()

    #c.blockchain.address.get_balance(23)




#!/usr/bin/env python3 
#
#
import bottom, random, time, asyncio
from .svr_info import ServerInfo
import logging

logger = logging.getLogger(__name__)

class IrcListener(bottom.Client):
    def __init__(self, irc_nickname=None, irc_password=None, ssl=True):
        self.my_nick = irc_nickname or 'XC%d' % random.randint(1E11, 1E12)
        self.password = irc_password or None

        self.results = {}       # by hostname
        self.servers = set()
        self.all_done = asyncio.Event()

        super(IrcListener, self).__init__(host='irc.freenode.net', port=6697 if ssl else 6667, ssl=ssl)

        # setup event handling
        self.on('CLIENT_CONNECT', self.connected)
        self.on('PING', self.keepalive)
        self.on('JOIN', self.joined)
        self.on('RPL_NAMREPLY', self.got_users)
        self.on('RPL_WHOREPLY', self.got_who_reply)
        self.on("client_disconnect", self.reconnect)
        self.on('RPL_ENDOFNAMES', self.got_end_of_names)

    async def collect_data(self):
        # start it process
        self.loop.create_task(self.connect())

        # wait until done
        await self.all_done.wait()

        # return the results
        return self.results

    def connected(self, **kwargs):
        logger.debug("Connected")
        self.send('NICK', nick=self.my_nick)
        self.send('USER', user=self.my_nick, realname='Connectrum Client')
        # long delay here as it does an failing Ident probe (10 seconds min)
        self.send('JOIN', channel='#electrum')
        #self.send('WHO', mask='E_*')

    def keepalive(self, message, **kwargs):
        self.send('PONG', message=message)

    async def joined(self, nick=None, **kwargs):
        # happens when we or someone else joins the channel
        # seem to take 10 seconds or longer for me to join
        logger.debug('Joined: %r' % kwargs)

        if nick != self.my_nick:
            await self.add_server(nick)

    async def got_who_reply(self, nick=None, real_name=None, **kws):
        '''
            Server replied to one of our WHO requests, with details.
        '''
        #logger.debug('who reply: %r' % kws)

        nick = nick[2:] if nick[0:2] == 'E_' else nick
        host, ports = real_name.split(' ', 1)

        self.servers.remove(nick)

        logger.debug("Found: '%s' at %s with port list: %s",nick, host, ports)
        self.results[host.lower()] = ServerInfo(nick, host, ports)

        if not self.servers:
            self.all_done.set()

    async def got_users(self, users=[], **kws):
        # After successful join to channel, we are given a list of 
        # users on the channel. Happens a few times for busy channels.
        logger.debug('Got %d (more) users in channel', len(users))

        for nick in users:
            await self.add_server(nick)

    async def add_server(self, nick):
        # ignore everyone but electrum servers
        if nick.startswith('E_'):
            self.servers.add(nick[2:])

    async def who_worker(self):
        # Fetch details on each Electrum server nick we see
        logger.debug('who task starts')
        copy = self.servers.copy()
        for nn in copy:
            logger.debug('do WHO for: ' + nn)
            self.send('WHO', mask='E_'+nn)

        logger.debug('who task done')

    def got_end_of_names(self, *a, **k):
        logger.debug('Got all the user names')

        assert self.servers, "No one on channel!"

        # ask for details on all of those users
        self.loop.create_task(self.who_worker())


    async def reconnect(self, **kwargs):
        # Trigger an event that may cascade to a client_connect.
        # Don't continue until a client_connect occurs, which may be never.

        logger.warn("Disconnected (will reconnect)")

        # Note that we're not in a coroutine, so we don't have access
        # to await and asyncio.sleep
        time.sleep(3)

        # After this line we won't necessarily be connected.
        # We've simply scheduled the connect to happen in the future
        self.loop.create_task(self.connect())

        logger.debug("Reconnect scheduled.")


if __name__ == '__main__':


    import logging
    logging.getLogger('bottom').setLevel(logging.DEBUG)
    logging.getLogger('connectrum').setLevel(logging.DEBUG)
    logging.getLogger('asyncio').setLevel(logging.DEBUG)


    bot = IrcListener(ssl=False)
    bot.loop.set_debug(True)
    fut = bot.collect_data()
    #bot.loop.create_task(bot.connect())
    rv = bot.loop.run_until_complete(fut)

    print(rv)





# copied values from electrum source

# IDK, maybe?
ELECTRUM_VERSION = '2.6.4'  # version of the client package
PROTOCOL_VERSION = '0.10'   # protocol version requested

# note: 'v' and 'p' are effectively reserved as well.
PROTOCOL_CODES = dict(t='TCP (plaintext)', h='HTTP (plaintext)', s='SSL', g='Websocket')

# from electrum/lib/network.py at Jun/2016
#
DEFAULT_PORTS = { 't':50001, 's':50002, 'h':8081, 'g':8082}

BOOTSTRAP_SERVERS = {
    'erbium1.sytes.net': {'t':50001, 's':50002},
    'ecdsa.net': {'t':50001, 's':110},
    'electrum0.electricnewyear.net': {'t':50001, 's':50002},
    'VPS.hsmiths.com': {'t':50001, 's':50002},
    'ELECTRUM.jdubya.info': {'t':50001, 's':50002},
    'electrum.no-ip.org': {'t':50001, 's':50002, 'g':443},
    'us.electrum.be': DEFAULT_PORTS,
    'bitcoins.sk': {'t':50001, 's':50002},
    'electrum.petrkr.net': {'t':50001, 's':50002},
    'electrum.dragonzone.net': DEFAULT_PORTS,
    'Electrum.hsmiths.com': {'t':8080, 's':995},
    'electrum3.hachre.de': {'t':50001, 's':50002},
    'elec.luggs.co': {'t':80, 's':443},
    'btc.smsys.me': {'t':110, 's':995},
    'electrum.online': {'t':50001, 's':50002},
}






"""
Setup file

See https://packaging.python.org/tutorials/distributing-packages/

But basically:
    python3 setup.py sdist
    (that makes a new tgz in ./dist)
    gpg -u 5A2A5B10 --detach-sign -a dist/connectrum-XXX.tar.gz
    twine upload dist/connectrum-XXX.*
    git tag vXXXX -a "New release"
    git push --tags

"""
import os
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.md')).read()


def get_version():
    with open("connectrum/__init__.py") as f:
        for line in f:
            if line.startswith("__version__"):
                return eval(line.split("=")[-1])

REQUIREMENTS = [
    # none at this time
]

TEST_REQUIREMENTS = [
    'aiohttp',
    'pytest',
    'tox',
    'aiosocks',
    'aiohttp',
    'bottom>=1.0.2'
]

if __name__ == "__main__":
    setup(
        name='connectrum',
        python_requires='>=3.6.0',
        version=get_version(),
        description="asyncio-based Electrum client library",
        long_description=README,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries',
        ],
        author='Peter Gray',
        author_email='peter@coinkite.com',
        url='https://github.com/coinkite/connectrum',
        license='MIT',
        keywords='electrum bitcoin asnycio client',
        platforms='any',
        include_package_data=True,
        packages=find_packages(exclude=('testing', 'examples')),
        #data_files=['connectrum/servers.json'],
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + TEST_REQUIREMENTS,
    )



#
# Store information about servers. Filter and select based on their protocol support, etc.
#

# Runtime check for optional modules
from importlib import util as importutil

# Check if bottom is present.
if importutil.find_spec("bottom") is not None:
    have_bottom = True
else:
    have_bottom = False

import time, random, json
from .constants import DEFAULT_PORTS


class ServerInfo(dict):
    '''
        Information to be stored on a server. Originally based on IRC data published to a channel.

    '''
    FIELDS = ['nickname', 'hostname', 'ports', 'version', 'pruning_limit' ]

    def __init__(self, nickname_or_dict, hostname=None, ports=None,
                        version=None, pruning_limit=None, ip_addr=None):

        if not hostname and not ports:
            # promote a dict, or similar
            super(ServerInfo, self).__init__(nickname_or_dict)
            return

        self['nickname'] = nickname_or_dict or None
        self['hostname'] = hostname
        self['ip_addr'] = ip_addr or None

        # For 'ports', take
        # - a number (int), assumed to be TCP port, OR
        # - a list of codes, OR
        # - a string to be split apart.
        # Keep version and pruning limit separate
        #
        if isinstance(ports, int):
            ports = ['t%d' % ports]
        elif isinstance(ports, str):
            ports = ports.split()

        # check we don't have junk in the ports list
        for p in ports.copy():
            if p[0] == 'v':
                version = p[1:]
                ports.remove(p)
            elif p[0] == 'p':
                try:
                    pruning_limit = int(p[1:])
                except ValueError:
                    # ignore junk
                    pass
                ports.remove(p)

        assert ports, "Must have at least one port/protocol"

        self['ports'] = ports
        self['version'] = version
        self['pruning_limit'] = int(pruning_limit or 0)

    @classmethod
    def from_response(cls, response_list):
        # Create a list of servers based on data from response from Stratum.
        # Give this the response to: "server.peers.subscribe"
        #
        #     [...
        #      "91.63.237.12",
        #      "electrum3.hachre.de",
        #      [ "v1.0", "p10000", "t", "s" ]
        #       ...]
        #
        rv = []

        for params in response_list:
            ip_addr, hostname, ports = params

            if ip_addr == hostname:
                ip_addr = None

            rv.append(ServerInfo(None, hostname, ports, ip_addr=ip_addr))

        return rv

    @classmethod
    def from_dict(cls, d):
        n = d.pop('nickname', None)
        h = d.pop('hostname')
        p = d.pop('ports')
        rv = cls(n, h, p)
        rv.update(d)
        return rv


    @property
    def protocols(self):
        rv = set(i[0] for i in self['ports'])
        assert 'p' not in rv, 'pruning limit got in there'
        assert 'v' not in rv, 'version got in there'
        return rv

    @property
    def pruning_limit(self):
        return self.get('pruning_limit', 100)

    @property
    def hostname(self):
        return self.get('hostname')

    def get_port(self, for_protocol):
        '''
            Return (hostname, port number, ssl) pair for the protocol.
            Assuming only one port per host.
        '''
        assert len(for_protocol) == 1, "expect single letter code"

        use_ssl = for_protocol in ('s', 'g')

        if 'port' in self: return self['hostname'], int(self['port']), use_ssl

        rv = next(i for i in self['ports'] if i[0] == for_protocol)

        port = None
        if len(rv) >= 2:
            try:
                port = int(rv[1:])
            except:
                pass
        port = port or DEFAULT_PORTS[for_protocol]

        return self['hostname'], port, use_ssl

    @property
    def is_onion(self):
        return self['hostname'].lower().endswith('.onion')

    def select(self, protocol='s', is_onion=None, min_prune=0):
        # predicate function for selection based on features/properties
        return ((protocol in self.protocols)
                    and (self.is_onion == is_onion if is_onion is not None else True)
                    and (self.pruning_limit >= min_prune))

    def __repr__(self):
        return '<ServerInfo {hostname} nick={nickname} ports="{ports}" v={version} prune={pruning_limit}>'\
                                        .format(**self)

    def __str__(self):
        # used as a dict key in a few places.
        return self['hostname'].lower()

    def __hash__(self):
        # this one-line allows use as a set member, which is really handy!
        return hash(self['hostname'].lower())

class KnownServers(dict):
    '''
        Store a list of known servers and their port numbers, etc.

        - can add single entries
        - can read from a CSV for seeding/bootstrap
        - can read from IRC channel to find current hosts

        We are a dictionary, with key being the hostname (in lowercase) of the server.
    '''

    def from_json(self, fname):
        '''
            Read contents of a CSV containing a list of servers.
        '''
        with open(fname, 'rt') as fp:
            for row in json.load(fp):
                nn = ServerInfo.from_dict(row)
                self[str(nn)] = nn

    def from_irc(self, irc_nickname=None, irc_password=None):
        '''
            Connect to the IRC channel and find all servers presently connected.

            Slow; takes 30+ seconds but authoritative and current.

            OBSOLETE.
        '''
        if have_bottom:
            from .findall import IrcListener

            # connect and fetch current set of servers who are
            # on #electrum channel at freenode

            bot = IrcListener(irc_nickname=irc_nickname, irc_password=irc_password)
            results = bot.loop.run_until_complete(bot.collect_data())
            bot.loop.close()

            # merge by nick name
            self.update(results)
        else:
            return(False)

    def add_single(self, hostname, ports, nickname=None, **kws):
        '''
            Explicitly add a single entry.
            Hostname is a FQDN and ports is either a single int (assumed to be TCP port)
            or Electrum protocol/port number specification with spaces in between.
        '''
        nickname = nickname or hostname

        self[hostname.lower()] = ServerInfo(nickname, hostname, ports, **kws)

    def add_peer_response(self, response_list):
        # Update with response from Stratum (lacks the nickname value tho):
        #
        #      "91.63.237.12",
        #      "electrum3.hachre.de",
        #      [ "v1.0", "p10000", "t", "s" ]
        #
        additions = set()
        for params in response_list:
            ip_addr, hostname, ports = params

            if ip_addr == hostname:
                ip_addr = None

            g = self.get(hostname.lower())
            nickname = g['nickname'] if g else None

            here = ServerInfo(nickname, hostname, ports, ip_addr=ip_addr)
            self[str(here)] = here

            if not g:
                additions.add(str(here))

        return additions

    def save_json(self, fname='servers.json'):
        '''
            Write out to a CSV file.
        '''
        rows = sorted(self.keys())
        with open(fname, 'wt') as fp:
            json.dump([self[k] for k in rows], fp, indent=1)

    def dump(self):
        return '\n'.join(repr(i) for i in self.values())

    def select(self, **kws):
        '''
            Find all servers with indicated protocol support. Shuffled.

            Filter by TOR support, and pruning level.
        '''
        lst = [i for i in self.values() if i.select(**kws)]

        random.shuffle(lst)

        return lst


if __name__ == '__main__':

    ks = KnownServers()

    #ks.from_json('servers.json')
    ks.from_irc()

    #print (ks.dump())

    from constants import PROTOCOL_CODES

    print ("%3d: servers in total" % len(ks))

    for tor in [False, True]:
        for pp in PROTOCOL_CODES.keys():
            ll = ks.select(pp, is_onion=tor)
            print ("%3d: %s" % (len(ll), PROTOCOL_CODES[pp] + (' [TOR]' if tor else '')))

# EOF
